from utils import create_mockapp

import rio


async def test_default_values_arent_considered_explicitly_set():
    class SquareComponent(rio.Component):
        label: str

        def __init__(self, label: str, size: float = 5):
            super().__init__(width=size, height=size)

            self.label = label

        def build(self):
            return rio.Text(self.label, width=self.width, height=self.height)

    class RootComponent(rio.Component):
        text: str

        def build(self):
            square_component = SquareComponent(self.text, size=10)
            return rio.Container(square_component)

    async with create_mockapp(lambda: RootComponent("Hello")) as app:
        root_component = app.get_component(RootComponent)
        square_component = app.get_component(SquareComponent)

        # Create a new SquareComponent with the default size. Since we aren't
        # explicitly passing a size to the constructor, reconciliation should
        # keep the old size.
        root_component.text = "World"
        await root_component.force_refresh()

        assert square_component.label == "World"
        assert square_component.width == 10
        assert square_component.height == 10


async def test_reconcile_same_component_instance():
    def build():
        return rio.Container(rio.Text("Hello"))

    async with create_mockapp(build) as app:
        app.outgoing_messages.clear()

        root_component = app.get_component(rio.Container)
        await root_component.force_refresh()

        # Nothing changed, so there's no need to send any data to JS. But in
        # order to know that nothing changed, the framework would have to track
        # every individual attribute of every component. Since we forced the
        # root_component to refresh, it's reasonable to send that component's data to
        # JS.
        assert not app.outgoing_messages or app.last_updated_components == {
            root_component
        }


async def test_reconcile_not_dirty_high_level_component():
    # Situation:
    # HighLevelComponent1 contains HighLevelComponent2
    # HighLevelComponent2 contains LowLevelContainer
    # HighLevelComponent1 is rebuilt and changes the child of LowLevelContainer
    # -> LowLevelContainer is reconciled and dirty (because it has new children)
    # -> HighLevelComponent2 is reconciled but *not* dirty because its child was
    # reconciled
    # The end result is that there is a new component (the child of
    # LowLevelContainer), whose builder (HighLevelComponent2) is not "dirty". Make
    # sure the new component is initialized correctly despite this.
    class HighLevelComponent1(rio.Component):
        switch: bool = False

        def build(self):
            if self.switch:
                child = rio.Switch()
            else:
                child = rio.Text("hi")

            return HighLevelComponent2(rio.Column(child))

    class HighLevelComponent2(rio.Component):
        content: rio.Component

        def build(self):
            return self.content

    async with create_mockapp(HighLevelComponent1) as app:
        root_component = app.get_component(HighLevelComponent1)
        root_component.switch = True
        await app.refresh()

        assert any(
            isinstance(component, rio.Switch)
            for component in app.last_updated_components
        )


async def test_reconcile_unusual_types():
    class Container(rio.Component):
        def build(self) -> rio.Component:
            return CustomComponent(
                integer=4,
                text="bar",
                tuple=(2.0, rio.Text("baz")),
                byte_array=bytearray(b"foo"),
            )

    class CustomComponent(rio.Component):
        integer: int
        text: str
        tuple: tuple[float, rio.Component]
        byte_array: bytearray

        def build(self):
            return rio.Text(self.text)

    async with create_mockapp(Container) as app:
        root_component = app.get_component(Container)

        # As long as this doesn't crash, it's fine
        await root_component.force_refresh()


async def test_reconcile_by_key():
    class Toggler(rio.Component):
        toggle: bool = False

        def build(self):
            if not self.toggle:
                return rio.Text("Hello", key="foo")
            else:
                return rio.Container(rio.Text("World", key="foo"))

    async with create_mockapp(Toggler) as app:
        root_component = app.get_component(Toggler)
        text = app.get_component(rio.Text)

        root_component.toggle = True
        await app.refresh()

        assert text.text == "World"


async def test_key_prevents_structural_match():
    class Toggler(rio.Component):
        toggle: bool = False

        def build(self):
            if not self.toggle:
                return rio.Text("Hello")
            else:
                return rio.Text("World", key="foo")

    async with create_mockapp(Toggler) as app:
        root_component = app.get_component(Toggler)
        text = app.get_component(rio.Text)

        root_component.toggle = True
        await app.refresh()

        assert text.text == "Hello"


async def test_key_interrupts_structure():
    class Toggler(rio.Component):
        key_: str = "abc"

        def build(self):
            return rio.Container(rio.Text(self.key_), key=self.key_)

    async with create_mockapp(Toggler) as app:
        root_component = app.get_component(Toggler)
        text = app.get_component(rio.Text)

        root_component.key_ = "123"
        await app.refresh()

        # The container's key changed, so even though the structure is the same,
        # the old Text component should be unchanged.
        assert text.text == "abc"


async def test_structural_matching_inside_keyed_component():
    class Toggler(rio.Component):
        toggle: bool = False

        def build(self):
            if not self.toggle:
                return rio.Row(rio.Container(rio.Text("A"), key="foo"))
            else:
                return rio.Row(
                    rio.Container(rio.Text("B")),
                    rio.Container(rio.Text("C"), key="foo"),
                )

    async with create_mockapp(Toggler) as app:
        root_component = app.get_component(Toggler)
        text = app.get_component(rio.Text)

        root_component.toggle = True
        await app.refresh()

        # The container with key "foo" has moved. Make sure the structure inside
        # of it was reconciled correctly.
        assert text.text == "C"


async def test_key_matching_inside_keyed_component():
    class Toggler(rio.Component):
        toggle: bool = False

        def build(self):
            if not self.toggle:
                return rio.Container(
                    rio.Row(
                        rio.Container(rio.Text("A"), key="container"),
                        key="row",
                    ),
                )
            else:
                return rio.Row(
                    rio.Text("B"),
                    rio.Container(rio.Text("C"), key="container"),
                    key="row",
                )

    async with create_mockapp(Toggler) as app:
        root_component = app.get_component(Toggler)
        text = app.get_component(rio.Text)

        root_component.toggle = True
        await app.refresh()

        # The container with key "foo" has moved. Make sure the structure inside
        # of it was reconciled correctly.
        assert text.text == "C"


async def test_same_key_on_different_component_type():
    class ComponentWithText(rio.Component):
        text: str

        def build(self) -> rio.Component:
            return rio.Text(self.text)

    class Toggler(rio.Component):
        toggle: bool = False

        def build(self):
            if not self.toggle:
                return rio.Text("Hello", key="foo")
            else:
                return ComponentWithText("World", key="foo")

    async with create_mockapp(Toggler) as app:
        root_component = app.get_component(Toggler)
        text = app.get_component(rio.Text)

        root_component.toggle = True
        await app.refresh()

        assert text.text == "Hello"


async def test_text_reconciliation():
    class RootComponent(rio.Component):
        text: str = "foo"

        def build(self) -> rio.Component:
            return rio.Text(self.text)

    async with create_mockapp(RootComponent) as app:
        root = app.get_component(RootComponent)
        text = app.get_component(rio.Text)

        root.text = "bar"
        await app.refresh()

        assert text.text == root.text


async def test_grid_reconciliation():
    class RootComponent(rio.Component):
        num_rows: int = 1

        def build(self) -> rio.Component:
            rows = [[rio.Text(f"Row {n}")] for n in range(self.num_rows)]
            return rio.Grid(*rows)

    async with create_mockapp(RootComponent) as app:
        root = app.get_component(RootComponent)
        grid = app.get_component(rio.Grid)

        root.num_rows += 1
        await app.refresh()

        assert {root, grid} < app.last_updated_components
        assert len(grid._children) == root.num_rows


async def test_margin_reconciliation():
    class RootComponent(rio.Component):
        switch: bool = True

        def build(self) -> rio.Component:
            if self.switch:
                return rio.Column(*[rio.Text("hi") for _ in range(7)])
            else:
                return rio.Column(
                    rio.Text("hi", margin_left=1),
                    rio.Text("hi", margin_right=1),
                    rio.Text("hi", margin_top=1),
                    rio.Text("hi", margin_bottom=1),
                    rio.Text("hi", margin_x=1),
                    rio.Text("hi", margin_y=1),
                    rio.Text("hi", margin=1),
                )

    async with create_mockapp(RootComponent) as app:
        root = app.get_component(RootComponent)
        texts = list(app.get_components(rio.Text))

        root.switch = False
        await app.refresh()

        assert texts[0].margin_left == 1
        assert texts[1].margin_right == 1
        assert texts[2].margin_top == 1
        assert texts[3].margin_bottom == 1

        assert texts[4].margin_x == 1
        assert texts[4].margin_left == texts[4].margin_right == None

        assert texts[5].margin_y == 1
        assert texts[5].margin_top == texts[5].margin_bottom == None

        assert texts[6].margin == 1
        assert (
            texts[6].margin_left
            == texts[6].margin_right
            == texts[6].margin_top
            == texts[6].margin_bottom
            == None
        )


async def test_reconcile_instance_with_itself():
    # There used to be a bug where, when a widget was reconciled with itself, it
    # was removed from `Session._dirty_components`. So any changes in that
    # widget weren't sent to the frontend.

    class Container(rio.Component):
        child: rio.Component

        def build(self) -> rio.Component:
            return self.child

    def build() -> rio.Component:
        return Container(rio.Text("foo"))

    async with create_mockapp(build, use_ordered_dirty_set=True) as app:
        container = app.get_component(Container)
        child = app.get_component(rio.Text)

        # Change the child's state and make its parent rebuild
        child.text = "bar"
        container.child = child

        # In order for the bug to occur, the parent has to be rebuilt before the
        # child
        assert list(app.session._dirty_components) == [child, container]
        await app.refresh()

        assert app.last_updated_components == {child, container}



from dataclasses import KW_ONLY, dataclass
from typing import *  # type: ignore

# https://github.com/forgedsoftware/measurementcommon/blob/master/systems.json


SI_PREFIXES = (
    (24, "yotta", "Y"),
    (21, "zetta", "Z"),
    (18, "exa", "E"),
    (15, "peta", "P"),
    (12, "tera", "T"),
    (9, "giga", "G"),
    (6, "mega", "M"),
    (3, "kilo", "k"),
    (2, "hecto", "h"),
    (1, "deca", "da"),
    (-1, "deci", "d"),
    (-2, "centi", "c"),
    (-3, "milli", "m"),
    (-6, "micro", "μ"),
    (-9, "nano", "n"),
    (-12, "pico", "p"),
    (-15, "femto", "f"),
    (-18, "atto", "a"),
    (-21, "zepto", "z"),
    (-24, "yocto", "y"),
)


INFORMATION_PREFIXES = (
    (10, "kibi", "Ki"),
    (20, "mebi", "Mi"),
    (30, "gibi", "Gi"),
    (40, "tebi", "Ti"),
    (50, "pebi", "Pi"),
    (60, "exbi", "Ei"),
    (70, "zebi", "Zi"),
    (80, "yobi", "Yi"),
)


ALL_PREFIXES = SI_PREFIXES + INFORMATION_PREFIXES


@dataclass(frozen=True)
class Dimension:
    names: Sequence[str]
    _: KW_ONLY
    ampere: int = 0
    candela: int = 0
    kelvin: int = 0
    kilogram: int = 0
    metre: int = 0
    mole: int = 0
    second: int = 0
    bytes: int = 0

    def __mul__(self, other: "Dimension") -> "Dimension":
        return Dimension(
            tuple(),
            ampere=self.ampere + other.ampere,
            candela=self.candela + other.candela,
            kelvin=self.kelvin + other.kelvin,
            kilogram=self.kilogram + other.kilogram,
            metre=self.metre + other.metre,
            mole=self.mole + other.mole,
            second=self.second + other.second,
            bytes=self.bytes + other.bytes,
        )


# Base dimensions
ELECTRICAL_CURRENT = Dimension(("electrical current", "current"), ampere=1)
LUMINOSITY = Dimension(("luminosity",), candela=1)
TEMPERATURE = Dimension(("temperature",), kelvin=1)
MASS = Dimension(("mass", "weight", "tonnage"), kilogram=1)
LENGTH = Dimension(("length", "distance"), metre=1)
AMOUNT_OF_SUBSTANCE = Dimension(("amount of substance",), mole=1)
TIME = Dimension(("time", "duration"), second=1)
FILE_SIZE = Dimension(("file size",), bytes=1)

# Common derived dimensions
AREA = Dimension(("area",), metre=2)
VOLUME = Dimension(("volume",), metre=3)
SPEED = Dimension(("speed",), metre=1, second=-1)
ACCELERATION = Dimension(("acceleration",), metre=1, second=-2)
JERK = Dimension(("jerk",), metre=1, second=-3)
FORCE = Dimension(("force",), kilogram=1, metre=1, second=-2)
PRESSURE = Dimension(("pressure",), kilogram=1, metre=-1, second=-2)
ENERGY = Dimension(("energy",), kilogram=1, metre=2, second=-2)
POWER = Dimension(("power",), kilogram=1, metre=2, second=-3)
MOMENTUM = Dimension(("momentum",), kilogram=1, metre=2, second=-1)
FREQUENCY = Dimension(("frequency",), second=-1)


@dataclass(frozen=True)
class Unit:
    dimension: Dimension
    base_multiple: float
    prefixes: Sequence[tuple[int, str, str]]


AMPERE = Unit(ELECTRICAL_CURRENT, 1, SI_PREFIXES)
CANDELA = Unit(LUMINOSITY, 1, SI_PREFIXES)
KELVIN = Unit(TEMPERATURE, 1, SI_PREFIXES)
GRAM = Unit(MASS, 1e-3, SI_PREFIXES)
METRE = Unit(LENGTH, 1, SI_PREFIXES)
MOLE = Unit(AMOUNT_OF_SUBSTANCE, 1, SI_PREFIXES)
SECOND = Unit(TIME, 1, SI_PREFIXES)
BIT = Unit(FILE_SIZE, 1 / 8, INFORMATION_PREFIXES)
BYTE = Unit(FILE_SIZE, 1, INFORMATION_PREFIXES)


METRIC_TONNE = Unit(MASS, 1e3, SI_PREFIXES)
US_TON = Unit(MASS, 907.18474, ())
POUND = Unit(MASS, 0.45359237, ())
OUNCE = Unit(MASS, 0.028349523125, ())
STONE = Unit(MASS, 6.35029318, ())
GRAIN = Unit(MASS, 0.00006479891, ())
CARAT = Unit(MASS, 0.0002, ())

US_FOOT = Unit(LENGTH, 0.3048, ())
US_INCH = Unit(LENGTH, 0.0254, ())

US_MILE = Unit(LENGTH, 1609.344, ())
NAUTICAL_MILE = Unit(LENGTH, 1852, ())
LIGHT_YEAR = Unit(LENGTH, 9.4607304725808e15, ())
PARSEC = Unit(LENGTH, 3.0856775814914e16, ())
ANGSTROM = Unit(LENGTH, 1e-10, ())
ASTRONOMICAL_UNIT = Unit(LENGTH, 1.495978707e11, ())



# fake module that resides in my lib folder and
# imports the actual implementation

from pathlib import Path

here = Path(__file__).absolute().parent
name = here.stem


import sys

sys.path.insert(0, str(here))
del sys.modules[name]
module = sys.modules[name] = __import__(name)
del sys.path[0]

del Path, here, name, sys
globals().update(module.__dict__)



from typing import cast

from utils import create_mockapp

import rio
from rio.state_properties import PleaseTurnThisIntoAnAttributeBinding


class Parent(rio.Component):
    text: str = ""

    def build(self):
        return rio.Text(self.bind().text)


class Grandparent(rio.Component):
    text: str = ""

    def build(self):
        return Parent(self.bind().text)


async def test_bindings_arent_created_too_early():
    # There was a time when attribute bindings were created in `Component.__init__`,
    # thus skipping any properties that were only assigned later.
    class IHaveACustomInit(rio.Component):
        text: str

        def __init__(self, *args, text: str, **kwargs):
            super().__init__(*args, **kwargs)

            # `Component.__init__`` has already run, but we haven't assigned
            # `self.text` yet. Do it now and assert that it still becomes a
            # attribute binding.
            self.text = text

        def build(self) -> rio.Component:
            return rio.Text(self.text)

    class Container(rio.Component):
        text: str = "hi"

        def build(self) -> rio.Component:
            return IHaveACustomInit(text=self.bind().text)

    async with create_mockapp(Container) as app:
        root_component = app.get_component(Container)
        child_component = app.get_component(IHaveACustomInit)

        assert child_component.text == "hi"

        root_component.text = "bye"
        assert child_component.text == "bye"


async def test_init_receives_attribute_bindings_as_input():
    # For a while we considered initializing attribute bindings before calling a
    # component's `__init__` and passing the values of the bindings as arguments
    # into `__init__`. But ultimately we decided against it, because some
    # components may want to use state properties/bindings in their __init__. So
    # make sure the `__init__` actually receives a
    # `PleaseTurnThisIntoAStateBinding` object as input.
    size_value = None

    class Square(rio.Component):
        def __init__(self, size: float):
            nonlocal size_value
            size_value = size

            super().__init__(width=size, height=size)

        def build(self) -> rio.Component:
            return rio.Text("hi", width=self.width, height=self.height)

    class Container(rio.Component):
        size: float

        def build(self) -> rio.Component:
            return Square(self.bind().size)

    async with create_mockapp(lambda: Container(7)):
        assert isinstance(size_value, PleaseTurnThisIntoAnAttributeBinding)


async def test_binding_assignment_on_child():
    async with create_mockapp(Parent) as app:
        root_component = app.get_component(Parent)
        text_component = app.get_build_output(root_component, rio.Text)

        assert not app.dirty_components

        text_component.text = "Hello"

        assert app.dirty_components == {root_component, text_component}
        assert root_component.text == "Hello"
        assert text_component.text == "Hello"


async def test_binding_assignment_on_parent():
    async with create_mockapp(Parent) as app:
        root_component = app.get_component(Parent)
        text_component = app.get_build_output(root_component)

        assert not app.dirty_components

        root_component.text = "Hello"

        assert app.dirty_components == {root_component, text_component}
        assert root_component.text == "Hello"
        assert text_component.text == "Hello"


async def test_binding_assignment_on_sibling():
    class Root(rio.Component):
        text: str = ""

        def build(self):
            return rio.Column(
                rio.Text(self.bind().text),
                rio.Text(self.bind().text),
            )

    async with create_mockapp(Root) as app:
        root_component = app.get_component(Root)
        text1, text2 = cast(
            list[rio.Text],
            app.get_build_output(root_component, rio.Column).children,
        )

        assert not app.dirty_components

        text1.text = "Hello"

        assert app.dirty_components == {root_component, text1, text2}
        assert root_component.text == "Hello"
        assert text1.text == "Hello"
        assert text2.text == "Hello"


async def test_binding_assignment_on_grandchild():
    async with create_mockapp(Grandparent) as app:
        root_component = app.get_component(Grandparent)
        parent = cast(Parent, app.get_build_output(root_component))
        text_component: rio.Text = app.get_build_output(parent)

        assert not app.dirty_components

        text_component.text = "Hello"

        assert app.dirty_components == {root_component, parent, text_component}
        assert root_component.text == "Hello"
        assert parent.text == "Hello"
        assert text_component.text == "Hello"


async def test_binding_assignment_on_middle():
    async with create_mockapp(Grandparent) as app:
        root_component = app.get_component(Grandparent)
        parent: Parent = app.get_build_output(root_component)
        text_component: rio.Text = app.get_build_output(parent)

        assert not app.dirty_components

        parent.text = "Hello"

        assert app.dirty_components == {root_component, parent, text_component}
        assert root_component.text == "Hello"
        assert parent.text == "Hello"
        assert text_component.text == "Hello"


async def test_binding_assignment_on_child_after_reconciliation():
    async with create_mockapp(Parent) as app:
        root_component = app.get_component(Parent)
        text_component: rio.Text = app.get_build_output(root_component)

        assert not app.dirty_components

        # Rebuild the root component, which reconciles the child
        await root_component.force_refresh()

        text_component.text = "Hello"

        assert app.dirty_components == {root_component, text_component}
        assert root_component.text == "Hello"
        assert text_component.text == "Hello"


async def test_binding_assignment_on_parent_after_reconciliation():
    async with create_mockapp(Parent) as app:
        root_component = app.get_component(Parent)
        text_component: rio.Text = app.get_build_output(root_component)

        assert not app.dirty_components

        # Rebuild the root component, which reconciles the child
        await root_component.force_refresh()

        root_component.text = "Hello"

        assert app.dirty_components == {root_component, text_component}
        assert root_component.text == "Hello"
        assert text_component.text == "Hello"


async def test_binding_assignment_on_sibling_after_reconciliation():
    class Root(rio.Component):
        text: str = ""

        def build(self):
            return rio.Column(
                rio.Text(self.bind().text),
                rio.Text(self.bind().text),
            )

    async with create_mockapp(Root) as app:
        root_component = app.get_component(Root)
        text1, text2 = app.get_build_output(root_component).children

        assert not app.dirty_components

        # Rebuild the root component, which reconciles the children
        await root_component.force_refresh()

        text1.text = "Hello"

        assert app.dirty_components == {root_component, text1, text2}
        assert root_component.text == "Hello"
        assert text1.text == "Hello"
        assert text2.text == "Hello"


async def test_binding_assignment_on_grandchild_after_reconciliation():
    async with create_mockapp(Grandparent) as app:
        root_component = app.get_component(Grandparent)
        parent: Parent = app.get_build_output(root_component)
        text_component: rio.Text = app.get_build_output(parent)

        assert not app.dirty_components

        # Rebuild the root component, which reconciles the child
        await root_component.force_refresh()

        text_component.text = "Hello"

        assert app.dirty_components == {root_component, parent, text_component}
        assert root_component.text == "Hello"
        assert parent.text == "Hello"
        assert text_component.text == "Hello"


async def test_binding_assignment_on_middle_after_reconciliation():
    async with create_mockapp(Grandparent) as app:
        root_component = app.get_component(Grandparent)
        parent: Parent = app.get_build_output(root_component)
        text_component: rio.Text = app.get_build_output(parent)

        assert not app.dirty_components

        # Rebuild the root component, which reconciles the child
        await root_component.force_refresh()

        parent.text = "Hello"

        assert app.dirty_components == {root_component, parent, text_component}
        assert root_component.text == "Hello"
        assert parent.text == "Hello"
        assert text_component.text == "Hello"



import subprocess
import sys

import requests
import revel

import rio


def main() -> None:
    revel.print("Running sanity checks...")

    # Make sure we're on the correct branch
    process = subprocess.run(
        ["git", "branch", "--show-current"],
        check=True,
        capture_output=True,
        text=True,
    )
    if process.stdout.strip("\n") != "dev":
        revel.fatal("You must checkout the 'dev' branch")

    # Make sure there are no uncommitted changes
    process = subprocess.run(
        ["git", "status", "--porcelain", "--untracked-files=no"],
        check=True,
        capture_output=True,
        text=True,
    )
    if process.stdout.strip():
        revel.fatal("There are uncommitted changes")

    # Make sure the version number got bumped
    if rio.__version__ == get_latest_published_version():
        revel.fatal("You forgot to increment the version number")

    # Build the TS code
    #
    # Note: `shell=True` is required on Windows because `npm` is a `.cmd` file
    # and not a `.exe`
    subprocess.run(["npm", "run", "build"], shell=True, check=True)

    # Run the test suite
    if tests_pass():
        revel.print("Everything is in order. Publishing...")
    else:
        # revel.fatal("The test don't pass")
        if not revel.select_yes_no(
            "The test suite does not pass. Do you want to publish anyway?"
        ):
            sys.exit(1)

        revel.print("Publishing...")

    # Merge dev into main and push
    subprocess.run(["git", "fetch", ".", "dev:main"], check=True)
    subprocess.run(["git", "push", "-u", "origin", "main"], check=True)

    # Publish
    subprocess.run(["rye", "build", "--clean"], check=True)
    subprocess.run(["rye", "publish"], check=True)


def get_latest_published_version() -> str:
    response = requests.get(f"https://pypi.org/pypi/rio-ui/json")
    return response.json()["info"]["version"]


def tests_pass() -> bool:
    process = subprocess.run(["rye", "test", "--", "-x", "--disable-warnings"])
    return process.returncode == 0


if __name__ == "__main__":
    main()



from __future__ import annotations

from collections.abc import Iterator
from typing import Any, TypeVar, cast

from . import dataclass, session, user_settings_module


__all__ = ["SessionAttachments"]


T = TypeVar("T")


class SessionAttachments:
    def __init__(self, sess: session.Session):
        self._session = sess
        self._attachments: dict[type, object] = {}

    def __iter__(self) -> Iterator[object]:
        return iter(self._attachments.values())

    def __contains__(self, typ: type) -> bool:
        return typ in self._attachments

    def __getitem__(self, typ: type[T]) -> T:
        try:
            return self._attachments[typ]  # type: ignore
        except KeyError:
            raise KeyError(typ) from None

    def _add(self, value: object, synchronize: bool) -> None:
        cls = type(value)

        # If the value isn't a UserSettings instance, just assign it and we're
        # done
        if not isinstance(value, user_settings_module.UserSettings):
            self._attachments[cls] = value
            return

        # If a UserSettings object is already attached, unlink it from the
        # session
        try:
            old_value = cast(
                user_settings_module.UserSettings, self._attachments[cls]
            )
        except KeyError:
            pass
        else:
            old_value._rio_session_ = None

        # Save the new value with the rest of the attachments
        self._attachments[cls] = value

        # Link the new value to the session
        value._rio_session_ = self._session

        # Trigger a resync
        if synchronize:
            # Since only assignments make attributes dirty, operations like
            # `list.append()` go unnoticed. So to guarantee that everything is
            # saved, we'll explicitly mark all attributes as dirty.
            value._rio_dirty_attribute_names_ = (
                set(dataclass.all_class_fields(cls))
                - user_settings_module.UserSettings.__annotations__.keys()
            )

            self._session._save_settings_soon()

    def add(self, value: Any) -> None:
        self._add(value, synchronize=True)

    def remove(self, typ: type) -> None:
        # Remove the attachment, propagating any `KeyError`
        old_value = self._attachments.pop(typ)

        # User settings need special care
        if not isinstance(old_value, user_settings_module.UserSettings):
            return

        # Unlink the value from the session
        old_value._rio_session_ = None

        # Trigger a resync
        #
        # TODO: `_save_settings_soon` doesn't currently delete any settings
        self._session._save_settings_soon()



"""
This file reads all materials icons/symbols from their github repository and
packs them into a `.tar.xz` archive that can be used by rio as icon set.

The repository is expected to be available locally already - this script does
not clone it.
"""

import re
import tarfile
import tempfile
from pathlib import Path
from typing import *  # type: ignore
from xml.etree import ElementTree as ET

import revel
from revel import *  # type: ignore

import rio

# Configure: The name the resulting icon set will have
SET_NAME = "material"

# Configure: The path to the directory which should be scanned for SVG files.
#
# Repo URL: https://github.com/marella/material-symbols
INPUT_DIR = (
    rio.utils.PROJECT_ROOT_DIR
    / "thirdparty"
    / "material-symbols"
    / "svg"
    / "500"
    / "rounded"
)

# Configure: Any files in the input directory which match this pattern will be
# processed
INPUT_NAME_PATTERN = r"(.+).svg"

# Configure: The output file will be written into this directory as
# <SET_NAME>.tar.xz
OUTPUT_DIR = rio.utils.RIO_ASSETS_DIR / "icon-sets"

# For debugging: Stop after processing this many icons. Set to `None` for no
# limit
LIMIT = None


# Configure: Given the relative path to the icon, return the icon's name and
# optionally variant.
#
# If `None` is returned the file is skipped.
def name_from_icon_path(path: Path) -> tuple[str, str | None | None]:
    # Normalize the name
    name = path.stem
    name = name.replace("_", "-")
    assert all(c.isalnum() or c == "-" for c in name), path

    # See if this is a variant
    known_variants = [
        "fill",
    ]

    for variant in known_variants:
        if name.endswith(f"-{variant}"):
            name = name.removesuffix(f"-{variant}")
            break
    else:
        variant = None

    return name, variant


# == No changes should be required below this line ==


def main() -> None:
    # Find all files in the input directory
    print_chapter("Scanning files")

    if not INPUT_DIR.exists():
        fatal(f"The input directory [bold]{INPUT_DIR}[/bold] does not exist")

    in_files: list[Path] = []
    for path in INPUT_DIR.glob("**/*"):
        if path.is_file() and re.fullmatch(INPUT_NAME_PATTERN, path.name):
            in_files.append(path)

    print(f"Found {len(in_files)} file(s)")

    # Enforce an order so that the output is deterministic
    in_files.sort()

    # Apply the limit
    if LIMIT is not None:
        in_files = in_files[:LIMIT]
        print(f"Limiting to {LIMIT} file(s) (set `LIMIT` to `None` to disable)")

    # Process all files
    print_chapter("Processing files")
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir = Path(tmp_dir)

        with revel.ProgressBar(max=len(in_files), unit="count") as bar:
            for ii, file_path in enumerate(in_files):
                bar.progress = ii

                # Extract the name and variant of the icon. If this function returns
                # `None` the file is skipped.
                parsed = name_from_icon_path(file_path.relative_to(INPUT_DIR))

                if parsed is None:
                    print(f"{file_path.name} -> [bold]skipped[/bold]")
                    continue

                icon_name, icon_variant = parsed
                variant_suffix = (
                    "" if icon_variant is None else f"/{icon_variant}"
                )

                print(f"{file_path.name} -> {icon_name}{variant_suffix}")

                # Parse the SVG
                svg_str = file_path.read_text()
                tree = ET.fromstring(svg_str)

                # Strip the width / height if any
                if "width" in tree.attrib:
                    del tree.attrib["width"]

                if "height" in tree.attrib:
                    del tree.attrib["height"]

                # Determine the output path for this icon
                icon_out_path = tmp_dir

                if icon_variant is not None:
                    icon_out_path /= icon_variant

                icon_out_path /= f"{icon_name}.svg"

                # Suppress weird "ns0:" prefixes everywhere
                ET.register_namespace("", "http://www.w3.org/2000/svg")

                # Write the SVG
                icon_out_path.parent.mkdir(parents=True, exist_ok=True)

                with open(icon_out_path, "w") as f:
                    f.write(
                        ET.tostring(
                            tree,
                            encoding="unicode",
                            default_namespace=None,
                        )
                    )

        # Compress the temporary directory
        print_chapter("Compressing files")
        archive_path = OUTPUT_DIR / f"{SET_NAME}.tar.xz"

        with tarfile.open(archive_path, "w:xz") as out_file:
            out_file.add(tmp_dir, arcname="")

    print_chapter(None)
    print(
        f"[bold]Done![/] You can find the result at [bold]{OUTPUT_DIR.resolve()}/{SET_NAME}.tar.xz[/]"
    )


if __name__ == "__main__":
    main()



import re
import subprocess
import tempfile
from typing import *  # type: ignore

import black
import imy.docstrings
import pyright
import pytest

import rio

CODE_BLOCK_PATTERN = re.compile(r"```.*?\n(.*?)\n```", re.DOTALL)


def all_components() -> list[Type[rio.Component]]:
    """
    Iterates over all components that ship with Rio.
    """
    to_do: Iterable[Type[rio.Component]] = [rio.Component]
    result: list[Type[rio.Component]] = []

    while to_do:
        component = to_do.pop()
        result.append(component)
        to_do.extend(component.__subclasses__())

    return result


def get_code_blocks(comp: Type[rio.Component]) -> list[str]:
    """
    Returns a list of all code blocks in the docstring of a component.
    """
    docs = imy.docstrings.ClassDocs.from_class(comp)

    # No docs?
    if docs.details is None:
        return []

    # Find any contained code blocks
    result: list[str] = []
    for match in CODE_BLOCK_PATTERN.finditer(docs.details):
        block: str = match.group(0)

        assert block.startswith("```")
        assert block.endswith("```")

        # Split into language and source
        linebreak = block.index("\n")
        assert linebreak != -1
        first_line = block[3:linebreak].strip()
        block = block[linebreak + 1 : -3]

        # Make sure a language is specified
        assert first_line, "The code block has no language specified"

        result.append(block)

    return result


@pytest.mark.parametrize("comp", all_components())
def test_eval_code_block(comp: Type[rio.Component]) -> None:
    # Eval all code blocks and make sure they don't crash
    for source in get_code_blocks(comp):
        compile(source, "<string>", "exec")


@pytest.mark.parametrize("comp", all_components())
def test_code_block_is_formatted(comp: Type[rio.Component]) -> None:
    # Make sure all code blocks are formatted according to black
    for source in get_code_blocks(comp):
        formatted_source = black.format_str(source, mode=black.FileMode())
        assert source == formatted_source


def _pyright_check_source(source: str) -> tuple[int, int]:
    """
    Run pyright on the given source and return the number of errors and
    warnings.
    """
    with tempfile.NamedTemporaryFile(suffix=".py") as f:
        # Dump the source to a file, and implicitly import rio
        f.write("import rio\n".encode("utf-8"))
        f.write(source.encode("utf-8"))
        f.flush()

        # Run pyright
        proc = pyright.run(
            f.name,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        result_out = proc.stdout
        assert isinstance(result_out, bytes), type(result_out)
        result_out = result_out.decode()

        # Find the number of errors and warnings
        match = re.search(
            r"(\d+) error(s)?, (\d+) warning(s)?, (\d+) information",
            result_out,
        )
        assert match is not None, result_out
        return int(match.group(1)), int(match.group(3))


@pytest.mark.parametrize("comp", all_components())
def test_analyze_code_block(comp: Type[rio.Component]) -> None:
    # Make sure pyright is happy with all code blocks
    for source in get_code_blocks(comp):
        n_errors, n_warnings = _pyright_check_source(source)

        assert n_errors == 0, f"Found {n_errors} errors"
        assert n_warnings == 0, f"Found {n_warnings} warnings"



from __future__ import annotations

from typing import *  # type: ignore

import rio

from .. import components as comps


# <component>
class HomePage(rio.Component):
    """
    A sample page, containing a greeting and some testimonials.
    """

    def build(self) -> rio.Component:
        return rio.Column(
            rio.Markdown(
                """
# Buzzwordz Inc.!

Unleashing synergistic paradigms for unprecedented excellence since the day
after yesterday.
            """,
                width=60,
                align_x=0.5,
            ),
            rio.Row(
                comps.Testimonial(
                    "Disruptive Innovations Inc. is the vanguard of operational excellence and groundbreaking innovation.",
                    "Jane Doe",
                    "CTO, Synergistic Solutions LLC",
                ),
                comps.Testimonial(
                    "They blew my socks off, and I wasn't even wearing any!",
                    "Made Up Rick",
                    "CEO, Imaginary Industries",
                ),
                comps.Testimonial(
                    "The Quantum Synergy Paradigm is a game-changer! I've never seen anything like it.",
                    "John Doe",
                    "CEO, HyperTech Corp.",
                ),
                spacing=2,
                align_x=0.5,
            ),
            rio.Markdown(
                """
Contact us today to unlock your business's quantum potential and embark on a
journey of transformational growth and stratospheric success.
            """,
                width=60,
                align_x=0.5,
            ),
            spacing=2,
            width=60,
            align_x=0.5,
            align_y=0,
        )


# </component>



# This silences the following error:
#
# Cancelling an overlapped future failed
# future: <_OverlappedFuture pending cb=[BaseProactorEventLoop._start_serving.<locals>.loop() at C:\Python312\Lib\asyncio\proactor_events.py:841, Task.task_wakeup()]>
# Traceback (most recent call last):
#   File "C:\Python312\Lib\asyncio\windows_events.py", line 71, in _cancel_overlapped
#     self._ov.cancel()
# OSError: [WinError 6] The handle is invalid

import asyncio.windows_events


def _cancel_overlapped(self):
    if self._ov is None:
        return

    try:
        self._ov.cancel()
    except OSError:
        pass

    self._ov = None


asyncio.windows_events._OverlappedFuture._cancel_overlapped = _cancel_overlapped  # type: ignore



import asyncio
import json
import logging
import signal
import socket
import threading
import time
import webbrowser
from typing import *  # type: ignore

import revel
from revel import print

import rio.app_server
import rio.cli
import rio.icon_registry
import rio.snippets

from ... import utils
from ...debug.monkeypatches import apply_monkeypatches
from .. import nice_traceback, project
from . import (
    app_loading,
    file_watcher_worker,
    run_models,
    uvicorn_worker,
    webview_worker,
)

try:
    import webview  # type: ignore
except ImportError:
    if TYPE_CHECKING:
        import webview  # type: ignore
    else:
        webview = None


# There is a hilarious problem with `watchfiles`: When a change occurs,
# `watchfiles` logs that change. If Python is configured to log into the project
# directory, then `watchfiles` will pick up on that new log line, thus
# triggering another change. Infinite loop, here we come!
#
# -> STFU, watchfiles!
logging.getLogger("watchfiles").setLevel(logging.CRITICAL + 1)


LOGO_TEXT = r"""  _____  _
 |  __ \(_)
 | |__) |_  ___
 |  _  /| |/ _ \
 | | \ \| | (_) |
 |_|  \_\_|\___/"""


class Arbiter:
    """
    Serves as a unified interface to running apps, coordinating the various
    workers necessary to do so.
    """

    def __init__(
        self,
        *,
        proj: project.RioProject,
        port: int | None,
        public: bool,
        quiet: bool,
        debug_mode: bool,
        run_in_window: bool,
    ) -> None:
        assert not (
            run_in_window and public
        ), "Can't run in a local window over the network, wtf!?"

        self.proj = proj
        self.public = public
        self.quiet = quiet
        self.debug_mode = debug_mode
        self.run_in_window = run_in_window

        # Workers communicate with the `RunningApp` by pushing events into this
        # queue.
        self._event_queue: asyncio.Queue[run_models.Event] = asyncio.Queue()

        # If running, contains the various workers
        self._file_watcher_worker: (
            file_watcher_worker.FileWatcherWorker | None
        ) = None
        self._uvicorn_worker: uvicorn_worker.UvicornWorker | None = None
        self._webview_worker: webview_worker.WebViewWorker | None = None

        self._file_watcher_task: asyncio.Task[None] | None = None
        self._uvicorn_task: asyncio.Task[None] | None = None
        self._arbiter_task: asyncio.Task[None] | None = None

        # The app to use for creating apps. This keeps the theme consistent if
        # for-example the user's app crashes and then a mock-app is injected.
        self._app_theme: Union[rio.Theme, tuple[rio.Theme, rio.Theme]] = (
            rio.Theme.pair_from_colors()
        )

        # Prefer to consistently run on the same port, as that makes it easier
        # to connect to - this way old browser tabs don't get invalidated
        # constantly.
        if port is None:
            if utils.port_is_free(self._host, 8000):
                self.port = 8000
            else:
                self.port = utils.choose_free_port(self._host)
        else:
            self.port = port

        # Used by the stop function for thread safety
        self._stopping_lock = threading.Lock()

        # This future resolves once the app is ready to be shown in the webview.
        # Wait for it.
        self._server_is_ready = threading.Event()

        # Used to signal that the app should close
        self._stop_requested = threading.Event()

        # The mainloop used by the arbiter. This is set once asyncio is running.
        self._mainloop: asyncio.AbstractEventLoop | None = None

    def push_event(self, event: run_models.Event) -> None:
        """
        Pushes an event into the event queue. Threadsafe.
        """
        assert (
            self._mainloop is not None
        ), "Can't push events before asyncio is running"
        rio.cli._logger.debug(f"Pushing arbiter event `{event}`")
        self._mainloop.call_soon_threadsafe(self._event_queue.put_nowait, event)

    @property
    def _host(self) -> str:
        return "0.0.0.0" if self.public else "127.0.0.1"

    @property
    def url(self) -> str:
        if self.public:
            # Get the local IP
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(
                ("8.8.8.8", 80)
            )  # Doesn't send data, because UDP is connectionless
            local_ip = s.getsockname()[0]
            s.close()
        else:
            local_ip = "127.0.0.1"

        return f"http://{local_ip}:{self.port}"

    @property
    def running_tasks(self) -> Iterator[asyncio.Task[None]]:
        for task in (
            self._uvicorn_task,
            self._file_watcher_task,
            self._arbiter_task,
        ):
            if task is not None:
                yield task

    def stop(self, *, keyboard_interrupt: bool) -> None:
        """
        Stops the app. This function can safely be called more than once, and at
        any point in time.
        """
        # Already shutting down? Then do nothing
        if not self._stopping_lock.acquire(blocking=False):
            return

        rio.cli._logger.debug("Stopping the app")

        # Visualize KeyboardInterrupts
        if keyboard_interrupt:
            print()
            print("[yellow]Interrupted[/]")
        else:
            print("Stopping")

        # Release anyone waiting for events
        self._stop_requested.set()
        self._server_is_ready.set()

        # Stop any running workers
        assert self._mainloop is not None, "Mainloop isn't running!?"

        def cancel_all_tasks() -> None:
            rio.cli._logger.debug(
                "Cancelling all arbiter tasks because the app is stopping"
            )

            for task in self.running_tasks:
                task.cancel()

        self._mainloop.call_soon_threadsafe(cancel_all_tasks)

        # Stop the webview
        if self._webview_worker is not None:
            rio.cli._logger.debug(
                "Stopping the webview because the app is stopping"
            )
            self._webview_worker.request_stop()

    def try_load_app(self) -> tuple[rio.App, Exception | None]:
        """
        Tries to load the user's app. If it fails, a dummy app is created and
        returned, unless running in release mode.

        Returns the new app and the error that occurred, if any.
        """
        rio.cli._logger.debug("Trying to load the app")

        try:
            app = app_loading.load_user_app(self.proj)

        except app_loading.AppLoadError as err:
            # If running in release mode, no further attempts to load the app
            # will be made. This error is fatal.
            if not self.debug_mode:
                rio.cli._logger.critical(f"The app could not be loaded: {err}")
                revel.fatal(f'The app could not be loaded: "{err}"')

            # Otherwise create a placeholder app which displays the error
            # message
            return (
                app_loading.make_error_message_app(
                    err,
                    self.proj.project_directory,
                    self._app_theme,
                ),
                err,
            )

        # Remember the app's theme. If in the future a placeholder app is used,
        # this theme will be used for it.
        self._app_theme = app._theme

        return app, None

    def run(self) -> None:
        assert not self._stop_requested.is_set()
        assert not self._server_is_ready.is_set()

        # Handle keyboard interrupts. KeyboardInterrupt exceptions are very
        # annoying to handle in async code, so instead we'll use a signal
        # handler.
        signal.signal(
            signal.SIGINT, lambda *_: self.stop(keyboard_interrupt=True)
        )

        # Do as much work as possible in asyncio land
        asyncio_thread = threading.Thread(
            target=lambda: asyncio.run(
                self._run_async(),
            ),
            name="arbiter function of rio run",
        )
        asyncio_thread.start()

        # Wait for the http server to be ready
        rio.cli._logger.debug("Waiting for the http server to be ready")
        self._server_is_ready.wait()

        # Make sure the startup was successful
        if self._stop_requested.is_set():
            rio.cli._logger.debug(
                "Stopping the arbiter because a stop was requested before the startup sequence has finished"
            )
            asyncio_thread.join()
            return

        # The webview needs to be shown from the main thread. So, if running
        # inside of a window run the arbiter in a separate thread. Otherwise
        # just run it from this one.
        if self.run_in_window:
            self._webview_worker = webview_worker.WebViewWorker(
                push_event=self.push_event,
                debug_mode=self.debug_mode,
                url=self.url,
            )

            rio.cli._logger.debug("Starting the webview worker")
            self._webview_worker.run()

        # If not running in a webview, just wait
        else:
            rio.cli._logger.debug("Opening the browser")
            webbrowser.open(self.url)

            # Event.wait() blocks the SIGINT handler, so we must periodically
            # return to python land to react to keyboard interrupts.
            #
            # We could pass a timeout to `_stop_requested.wait()`, but
            # `time.sleep()` reacts to keyboard interrupts immediately, which is
            # preferable.
            while True:
                time.sleep(0.2)

                if self._stop_requested.is_set():
                    break

        # Make sure the main thread stays alive until the asyncio thread is
        # done. Otherwise we get weird errors like "Can't start thread during
        # interpreter shutdown" from asyncio.
        rio.cli._logger.debug(
            "The arbiter is waiting for the asyncio thread to finish"
        )
        asyncio_thread.join()

        rio.cli._logger.debug("Arbiter shutdown complete")

    async def _run_async(self) -> None:
        # Publish the task, so it can be cancelled
        self._arbiter_task = asyncio.current_task()

        # Make sure the webview module is available
        if self.run_in_window and webview is None:
            revel.fatal(
                "The `window` extra is required to run apps inside of a window."
                " Run `pip install rio-ui[[window]` to install it."
            )

        # Make sure the app is cleanly shut down, even if the arbiter crashes
        # for whichever reason.
        #
        # This loop has a bad case of hiding important exceptions. It would
        # really be nice to only shut down on intentional errors and at least
        # log the others.
        try:
            await self._run_async_inner()

        # Keyboard interrupts are already handled by the signal handler
        except KeyboardInterrupt:
            pass

        # The worker is expected to be cancelled when the app is stopped
        except asyncio.CancelledError:
            pass

        # Anything else is an unintentional crash
        except Exception as err:
            revel.error("The arbiter has crashed.")
            revel.error("This is a bug in Rio - please report it")
            print()
            revel.print(nice_traceback.format_exception_revel(err))

            rio.cli._logger.exception("The arbiter has crashed")

            self.stop(keyboard_interrupt=False)

        finally:
            # Wait until all the other tasks are done. If we return, then
            # `asyncio.run` will cancel anything that's still running.
            rio.cli._logger.debug("Waiting for all arbiter tasks to finish")

            for task in self.running_tasks:
                if task is self._arbiter_task:
                    continue

                try:
                    await task
                except asyncio.CancelledError:
                    pass

            # Then cancel all remaining tasks to ensure the program exits
            rio.cli._logger.debug("Cancelling all remaining asyncio tasks")

            for task in asyncio.all_tasks():
                if task is self._arbiter_task:
                    continue

                task.cancel()

    async def _run_async_inner(self) -> None:
        # Print some initial messages
        print()
        print(f"[bold green]{LOGO_TEXT}[/]")
        print()
        print()
        print("Starting...")

        # Expose the mainloop
        self._mainloop = asyncio.get_running_loop()

        # Make sure the chosen port is available
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            rio.cli._logger.debug(f"Checking if port {self.port} is available")

            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

            try:
                sock.bind((self._host, self.port))
            except OSError:
                revel.error(f"The port [bold]{self.port}[/] is already in use.")
                revel.error(f"Each port can only be used by one app at a time.")
                revel.error(
                    f"Try using another port, or let Rio choose one for you, by not specifying any port."
                )
                self.stop(keyboard_interrupt=False)
                return

            # If running in debug mode, install safeguards
            if self.debug_mode:
                rio.cli._logger.debug("Applying monkeypatches")
                apply_monkeypatches()

            # Try to load the app
            app, _ = self.try_load_app()

            # Start the file watcher
            if self.debug_mode:
                self._file_watcher_worker = (
                    file_watcher_worker.FileWatcherWorker(
                        push_event=self.push_event,
                        proj=self.proj,
                    )
                )
                self._file_watcher_task = asyncio.create_task(
                    self._file_watcher_worker.run(),
                    name="File watcher",
                )

            # Start the uvicorn worker
            uvicorn_is_ready_or_has_failed: asyncio.Future[None] = (
                asyncio.Future()
            )

            self._uvicorn_worker = uvicorn_worker.UvicornWorker(
                push_event=self.push_event,
                app=app,
                socket=sock,
                quiet=self.quiet,
                debug_mode=self.debug_mode,
                run_in_window=self.run_in_window,
                on_server_is_ready_or_failed=uvicorn_is_ready_or_has_failed,
            )

            self._uvicorn_task = asyncio.create_task(
                self._uvicorn_worker.run(),
                name="Uvicorn",
            )

            # Wait for the server to be ready or fails to start
            rio.cli._logger.debug("Waiting for uvicorn to be ready or to fail")
            await uvicorn_is_ready_or_has_failed

            # Let everyone else know that the server is ready
            rio.cli._logger.debug(
                "App startup complete and ready for connections"
            )
            self._server_is_ready.set()

            # The app has just successfully started. Inform the user
            if self.debug_mode:
                revel.warning("Debug mode is enabled.")
                revel.warning(
                    "Debug mode includes helpful tools for development, but is slower and disables some safety checks. Never use it in production!"
                )
                revel.warning("Run with `--release` to disable debug mode.")
                print()

            if self.public:
                revel.warning(
                    f"Running in public mode. All devices on your network can access the app."
                )
                revel.warning(
                    f"Only run in public mode if you trust your network!"
                )
                revel.warning(
                    f"Run without `--public` to limit access to this device."
                )
                print()
            elif not self.run_in_window:
                print(
                    f"[dim]Running in [/]local[dim] mode. Only this device can access the app.[/]"
                )

            if not self.run_in_window:
                revel.success(f"The app is running at [bold]{self.url}[/]")
                print()

            # Keep track when the app was last reloaded
            last_reload_started_at = -1

            # Listen for events and react to them
            while True:
                event = await self._event_queue.get()
                rio.cli._logger.debug(f"Received arbiter event `{event}`")

                # A file has changed
                if isinstance(event, run_models.FileChanged):
                    # Ignore events that happened before the last reload started
                    if event.timestamp_nanoseconds < last_reload_started_at:
                        continue

                    # Display to the user that a reload is happening
                    rel_path = event.path_to_file.relative_to(
                        self.proj.project_directory
                    )
                    print()
                    print(f"[bold]{rel_path}[/] has changed -> Reloading")

                    # Update the timestamp
                    last_reload_started_at = time.monotonic_ns()

                    # If the `rio.toml` has changed, reload the entire project,
                    # not just the app
                    if event.path_to_file == self.proj.rio_toml_path:
                        rio.cli._logger.info(
                            "Reloading project configuration from `rio.toml`"
                        )
                        print("Reloading project configuration from `rio.toml`")
                        self.proj.discard_changes_and_reload()

                    # Restart the app
                    await self._restart_app()

                # Somebody requested that the app should stop
                elif isinstance(event, run_models.StopRequested):
                    self.stop(keyboard_interrupt=False)
                    break

                # ???
                else:
                    raise NotImplementedError(f'Unknown event "{event}"')

    def _spawn_traceback_popups(self, err: Union[str, BaseException]) -> None:
        """
        Displays a popup with the traceback in the rio UI.
        """
        assert self._uvicorn_worker is not None
        assert self._uvicorn_worker.app_server is not None

        rio.cli._logger.debug("Spawning traceback popups")

        popup_html = app_loading.make_traceback_html(
            err=err,
            project_directory=self.proj.project_directory,
        )

        for (
            session
        ) in self._uvicorn_worker.app_server._active_session_tokens.values():
            self._evaluate_javascript_in_session_if_connected(
                session,
                f"""
// Override the popup with the traceback message
let popup = document.querySelector(".rio-connection-lost-popup");
popup.innerHTML = {json.dumps(popup_html)};

// Spawn the popup
window.setConnectionLostPopupVisible(true);
""",
            )

    async def _restart_app(self) -> None:
        assert self._uvicorn_worker is not None

        rio.cli._logger.debug("Beginning app restart sequence")

        app_server = self._uvicorn_worker.app_server
        assert app_server is not None

        # Shut down the current app. This happens automatically when the app
        # server shuts down, but since we're just swapping out the app, we have
        # to call it manually.
        # TODO: Shouldn't we close all the active sessions first? It's odd that
        # `on_app_close` is called while there are still active sessions.
        await app_server._call_on_app_close()

        # Load the user's app again
        new_app, loading_error = self.try_load_app()

        # Replace the app which is currently hosted by uvicorn
        self._uvicorn_worker.replace_app(new_app)
        revel.success("Ready")

        # There is a subtlety here. Sessions which have requested their
        # index.html, but aren't yet connected to the websocket cannot
        # receive messages. So `_spawn_traceback_popups` will skip them.
        # This is fine as long as `self._app_server.app` has already been
        # assigned, since this ensures that any new websocket connections
        # will receive the new app anyway.
        #
        # -> This MUST happen after the new app has already been injected
        if loading_error is not None:
            self._spawn_traceback_popups(loading_error)

        # The app has changed, but the uvicorn server is still the same. Because
        # of this, uvicorn won't call the `on_app_start` function - do it
        # manually.
        await app_server._call_on_app_start()

        # Tell all sessions to reconnect, and close old sessions
        rio.cli._logger.debug("Reloading all sessions")

        for sess in list(app_server._active_session_tokens.values()):
            # Tell the session to reload
            #
            # TODO: Should there be some waiting period here, so that the
            # session has time to save settings first and shut down in general?
            self._evaluate_javascript_in_session_if_connected(
                sess, "window.location.reload()"
            )

            # Close it
            asyncio.create_task(
                sess._close(close_remote_session=False),
                name=f'Close session "{sess._session_token}"',
            )

    def _evaluate_javascript_in_session_if_connected(
        self,
        session: rio.Session,
        javascript_source: str,
    ) -> None:
        """
        Runs the given javascript source in the given session. Does not
        wait for or return the result.

        Does nothing if the session hasn't fully connected yet.
        """
        # If this session isn't done connecting, just return
        if session._send_message is rio.session.dummy_send_message:
            return

        # Run the javascript in a task
        async def evaljs_as_coroutine() -> None:
            await session._evaluate_javascript(javascript_source)

        session.create_task(
            evaljs_as_coroutine(),
            name=f"Eval JS in session {session._session_token}",
        )



from __future__ import annotations

from dataclasses import KW_ONLY
from typing import Literal, final

from typing_extensions import Self

import rio

from .fundamental_component import FundamentalComponent

__all__ = ["FlowContainer"]


@final
class FlowContainer(FundamentalComponent):
    """
    A container that lays out its children in a horizontal or vertical flow.

    `FlowContainer` is a container that lays out its children in a horizontal or
    vertical flow. It is similar to `Container`, but allows you to specify
    spacing between the children.


    ## Attributes

    `children`: The components to place inside the container.

    `row_spacing`: The vertical spacing between the children.

    `column_spacing`: The horizontal spacing between the children.

    `justify`: The horizontal alignment of the children. Possible values are:
        `left`, `center`, `right`, `justified` and `grow`.


    ## Examples

    This minimal example will show a container with a horizontal flow:

    ```python
    rio.FlowContainer(
        rio.Text("Hello"),
        rio.Text("World!"),
        column_spacing=1,
    )
    ```
    """

    children: list[rio.Component]
    _: KW_ONLY
    column_spacing: float
    row_spacing: float
    justify: Literal["left", "center", "right", "justified", "grow"]

    def __init__(
        self,
        *children: rio.Component,
        row_spacing: float = 0.0,
        column_spacing: float = 0.0,
        justify: Literal[
            "left", "center", "right", "justified", "grow"
        ] = "center",
        key: str | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        width: float | Literal["natural", "grow"] = "natural",
        height: float | Literal["natural", "grow"] = "natural",
        align_x: float | None = None,
        align_y: float | None = None,
    ):
        assert isinstance(children, tuple), children

        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        self.children = list(children)
        self.column_spacing = column_spacing
        self.row_spacing = row_spacing
        self.justify = justify

    def add(self, child: rio.Component) -> Self:
        """
        Appends a child component.

        Appends a child component to the end and then returns the
        `FlowContainer`, which makes method chaining possible:

        ```python
        rio.FlowContainer().add(child1).add(child2)
        ```


        ## Parameters

        `child`: The child component to append.
        """
        self.children.append(child)
        return self


FlowContainer._unique_id = "FlowContainer-builtin"



import logging

_logger = logging.getLogger(__name__)


from typing import Literal

import introspection
import revel
from revel import *  # type: ignore

import rio.snippets

from . import project, project_setup, run_project

__all__ = [
    "app",
]

revel.GLOBAL_STYLES.add_alias("primary", ["cyan"])
revel.GLOBAL_STYLES.add_alias("bg-primary", ["bg-cyan"])

app = revel.App(
    nicename="Rio",
    command_name="rio",
    summary="An easy to use, web & app framework for Python",
    details="""
Rio is a framework for building reactive apps and websites in Python. It's
designed to be easy to use, and to get out of your way as much as possible.

This is the command line interface for Rio. You can use it to easily create new
projects, run them, and more.
""",
    version=rio.__version__,
)


@app.command(
    aliases={"init", "create"},
    summary="Create a new Rio project",
    details="""
The `new` command creates a new directory and populates it with the files needed
to start a new Rio project. You can optionally specify a template, in which case
the files from that template will be copied into the new project, allowing you
hit the ground running.
""",
    parameters=[
        revel.Parameter(
            "nicename",
            summary="Human-readable name for the new project",
            prompt="What should the project be called?",
        ),
        revel.Parameter(
            "type",
            summary="Whether to create a website or an app",
        ),
        revel.Parameter(
            "template",
            summary="Template to use for the new project",
        ),
    ],
)
def new(
    nicename: str,
    *,
    # Website is listed first to make it the default
    type: Literal["website", "app"],
    template: rio.snippets.AvailableTemplatesLiteral,
) -> None:
    project_setup.create_project(
        raw_name=nicename,
        type=type,
        template_name=template,
    )


@app.command(
    summary="Run the current project",
    parameters=[
        revel.Parameter(
            "release",
            summary="Switches between release and debug mode",
        ),
        revel.Parameter(
            "port",
            summary="Port to run the HTTP server on",
        ),
        revel.Parameter(
            "public",
            summary="Whether the app should be available on the local network, rather than just the local device",
        ),
        revel.Parameter(
            "quiet",
            summary="Suppresses HTTP logs and other noise",
        ),
    ],
    details="""
The `run` command runs the current project for debugging. If your project is a
website, it will be hosted on a local HTTP server, and you can view it in your
browser. If the project is a app, it will be displayed in a window instead.

Rio will constantly watch your project for changes, and automatically reload
the app or website when it detects a change. This makes it easy to iterate on
your project without having to manually restart it.

The `port` and `public` options are ignored if the project is an app, since
they only make sense for websites.
""",
)
def run(
    *,
    release: bool = False,
    port: int | None = None,
    public: bool = False,
    quiet: bool = True,
) -> None:
    with project.RioProject.try_locate_and_load() as proj:
        # Some options only make sense for websites
        if proj.app_type == "app":
            if port is not None:
                port = None
                warning(
                    "Ignoring the `port` option, since this project is not a website"
                )

            if public:
                public = False
                warning(
                    "Ignoring the `public` option, since this project is not a website"
                )

        # Running a project comes with considerable complexity. All of that is
        # crammed into classes.
        arbiter = run_project.Arbiter(
            proj=proj,
            port=port,
            public=public,
            quiet=quiet,
            debug_mode=not release,
            run_in_window=proj.app_type == "app",
        )
        arbiter.run()


@app.command(
    summary="Add a page or component to the project",
    parameters=[
        revel.Parameter(
            "what",
            summary="Whether to add a `page` or a `component`",
        ),
        revel.Parameter(
            "name",
            summary="The name of the new page or component",
        ),
    ],
    details="""
The `add` command adds a new page or component to your project. A python file
containing some template code will be created in the `pages` or `components`
folder of your project.
""",
)
def add(what: Literal["page", "component"], /, name: str) -> None:
    with project.RioProject.try_locate_and_load() as proj:
        module_path = proj.module_path
        if not module_path.is_dir():
            fatal(
                f"Cannot add {what}s to a single-file project. Please convert"
                f" your project into a package."
            )

        # Make sure the `pages` or `components` folder exists
        folder_path = module_path / (what + "s")
        folder_path.mkdir(exist_ok=True)

        # Create the new file
        name = name.strip().replace(" ", "_")
        class_name = introspection.convert_case(name, "pascal")

        file_name = introspection.convert_case(name, "snake")
        file_path = folder_path / (file_name + ".py")

        if file_path.exists():
            fatal(f"File {file_path.relative_to(module_path)} already exists")

        # Write the file content
        if what == "page":
            file_path.write_text(
                f"""from __future__ import annotations

from typing import *  # type: ignore

import rio

from .. import components as comps


class {class_name}(rio.Component):
    def build(self) -> rio.Component:
        return rio.Markdown(
            '''
## This is a Sample Page

Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor
incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis
nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.
Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu
fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in
culpa qui officia deserunt mollit anim id est laborum.
            '''
        )
"""
            )
        else:
            file_path.write_text(
                f"""from __future__ import annotations

from dataclasses import KW_ONLY, field
from typing import *  # type: ignore

import rio

from .. import components as comps


class {class_name}(rio.Component):
    example_state: str = "For demonstration purposes"

    def build(self) -> rio.Component:
        return rio.Text(self.example_state)
"""
            )

        # Import the module in the __init__.py
        init_py_path = file_path.with_name("__init__.py")
        try:
            init_py_code = init_py_path.read_text(encoding="utf8")
        except FileNotFoundError:
            init_py_code = ""
        init_py_code = (
            init_py_code.rstrip() + f"\nfrom .{file_name} import {class_name}\n"
        )
        init_py_path.write_text(init_py_code, encoding="utf8")

        success(
            f"New {what} created at {file_path.relative_to(proj.project_directory)}"
        )



from pathlib import Path
from typing import *  # type: ignore

import gitignore_parser

__all__ = [
    "RioIgnore",
]


class RioIgnore:
    """
    Parses `.rioignore` files and allows querying whether a given path should be
    ignored.

    The actual rule matching is left to `gitignore_parser`, this class is just a
    wrapper making it easier to use and extending it slightly.
    """

    def __init__(self, base_dir: Path):
        self._base_dir = base_dir.resolve()
        self._rules: list[gitignore_parser.IgnoreRule] = []

    def add_pattern(self, line: str) -> None:
        """
        Add a pattern to the ignore list.
        """
        # Parse the pattern
        line = line.rstrip("\n")
        pattern = gitignore_parser.rule_from_pattern(
            line,
            base_path=self._base_dir,
            source=(None, len(self._rules)),
        )

        # Empty line / comment
        if not pattern:
            return

        self._rules.append(pattern)

    def add_patterns_from_file(self, file_path: TextIO) -> None:
        """
        Convenience method for adding all patterns from a file.
        """
        for line in file_path.readlines():
            self.add_pattern(line)

    def is_path_ignored(self, path: Path) -> bool:
        """
        Given a path, return whether the given path should be ignored, i.e.
        there is a rule matching it.
        """
        for rule in reversed(self._rules):
            if rule.match(path):
                return not rule.negation

        return False

    def is_explicitly_included(self, path: Path) -> bool:
        """
        Given a path, return whether there is an explicit exception for the
        file. A file is considered explicitly included if there is a rule
        of the form `!path`.
        """
        # Prepare the exact pattern which would match this path. This will also
        # find any paths which aren't in the basedirectory at all.
        try:
            pattern_should = f"!{path.relative_to(self._base_dir)}"
        except ValueError:
            return False

        # Compare all rules:
        for rule in self._rules:
            if rule.pattern.strip() == pattern_should:
                return True

        # No match
        return False



from __future__ import annotations

import dataclasses
import types
import weakref
from collections.abc import Callable
from typing import TYPE_CHECKING

import introspection.typing

from . import global_state

if TYPE_CHECKING:
    from .components import Component


__all__ = [
    "StateProperty",
    "AttributeBinding",
    "AttributeBindingMaker",
    "PleaseTurnThisIntoAnAttributeBinding",
]


class StateProperty:
    """
    StateProperties act like regular properties, with additional considerations:

    - When a state property is assigned to, the component owning it is marked as
      dirty in the session

    - State properties have the ability to share their value with other state
      property instances. If state property `A` is assigned to state property
      `B`, then `B` creates a `StateBinding` and any future access to `B` will
      be routed to `A` instead:

    ```python
    class Foo(Component):
        foo_text = "Hello"

        def build(self) -> Component:
            return Bar(bar_text=Foo.foo_text)  # Note `Foo` instead of `self`
    ```
    """

    def __init__(
        self,
        name: str,
        readonly: bool,
        raw_annotation: introspection.types.TypeAnnotation,
        module: types.ModuleType,
    ):
        self.name = name
        self.readonly = readonly

        self._module = module
        self._raw_annotation = raw_annotation
        self._resolved_annotation: introspection.types.TypeAnnotation

    def _annotation_as_string(self) -> str:
        if isinstance(self._raw_annotation, str):
            return self._raw_annotation

        return introspection.typing.annotation_to_string(self._raw_annotation)

    def __get__(
        self,
        instance: Component | None,
        owner: type | None = None,
    ) -> object:
        # If accessed through the class, rather than instance, return the
        # StateProperty itself
        if instance is None:
            return self

        # Otherwise get the value assigned to the property in the component
        # instance
        try:
            value = vars(instance)[self.name]
        except KeyError:
            raise AttributeError(self.name) from None

        # If the value is a binding return the binding's value
        if isinstance(value, AttributeBinding):
            return value.get_value()

        # Otherwise return the value
        return value

    def __set__(self, instance: Component, value: object) -> None:
        if self.readonly:
            cls_name = type(instance).__name__
            raise AttributeError(
                f"Cannot assign to readonly property {cls_name}.{self.name}"
            )

        assert not isinstance(
            value, StateProperty
        ), f"You're still using the old attribute binding syntax for {instance} {self.name}"

        instance._properties_assigned_after_creation_.add(self.name)

        # Look up the stored value
        instance_vars = vars(instance)
        try:
            local_value = instance_vars[self.name]
        except KeyError:
            # If no value is currently stored, that means this component is
            # currently being instantiated. We may have to create a state
            # binding.
            if isinstance(value, PleaseTurnThisIntoAnAttributeBinding):
                value = self._create_attribute_binding(instance, value)
        else:
            # If a value is already stored, that means this is a re-assignment.
            # Which further means it's an assignment outside of `__init__`.
            # Which is not a valid place to create a attribute binding.
            if isinstance(value, PleaseTurnThisIntoAnAttributeBinding):
                raise RuntimeError(
                    "Attribute bindings can only be created when calling the component constructor"
                )

            # Delegate to the binding if it exists
            if isinstance(local_value, AttributeBinding):
                local_value.set_value(value)
                return

        # Otherwise set the value directly and mark the component as dirty
        instance_vars[self.name] = value

        instance._session_._register_dirty_component(
            instance,
            include_children_recursively=False,
        )

    def _create_attribute_binding(
        self,
        component: Component,
        request: PleaseTurnThisIntoAnAttributeBinding,
    ) -> AttributeBinding:
        # In order to create a `StateBinding`, the creator's attribute must
        # also be a binding
        creator = global_state.currently_building_component
        creator_vars = vars(creator)

        parent_binding = creator_vars[request.state_property.name]

        if not isinstance(parent_binding, AttributeBinding):
            parent_binding = AttributeBinding(
                owning_component_weak=weakref.ref(creator),
                owning_property=self,
                is_root=True,
                parent=None,
                value=parent_binding,
                children=weakref.WeakSet(),
            )
            creator_vars[request.state_property.name] = parent_binding

        # Create the child binding
        child_binding = AttributeBinding(
            owning_component_weak=weakref.ref(component),
            owning_property=self,
            is_root=False,
            parent=parent_binding,
            value=None,
            children=weakref.WeakSet(),
        )
        parent_binding.children.add(child_binding)

        return child_binding

    def __repr__(self) -> str:
        return f"<{type(self).__name__} {self.name}>"


@dataclasses.dataclass(eq=False)
class AttributeBinding:
    # Weak reference to the component containing this binding
    owning_component_weak: Callable[[], Component | None]

    # The state property whose value this binding is
    owning_property: StateProperty

    # Each binding is either the root-most binding, or a child of another
    # binding. This value is True if this binding is the root.
    is_root: bool

    parent: AttributeBinding | None
    value: object | None

    children: weakref.WeakSet[AttributeBinding] = dataclasses.field(
        default_factory=weakref.WeakSet
    )

    def get_value(self) -> object:
        if self.is_root:
            return self.value

        assert self.parent is not None
        return self.parent.get_value()

    def set_value(self, value: object) -> None:
        # Delegate to the parent, if any
        if self.parent is not None:
            self.parent.set_value(value)
            return

        # Otherwise this is the root-most binding. Set the value
        self.value = value

        # Then recursively mark all children as dirty
        self.recursively_mark_children_as_dirty()

    def recursively_mark_children_as_dirty(self) -> None:
        to_do: list[AttributeBinding] = [self]

        while to_do:
            cur = to_do.pop()
            owning_component = cur.owning_component_weak()

            if owning_component is not None:
                owning_component._session_._register_dirty_component(
                    owning_component,
                    include_children_recursively=False,
                )

            to_do.extend(cur.children)


class AttributeBindingMaker:
    """
    Helper class returned by `Component.bind()`. Used to create attribute bindings.
    """

    def __init__(self, component: Component):
        self.component = component

    def __getattribute__(self, name: str) -> object:
        component = super().__getattribute__("__dict__")["component"]
        component_cls = type(component)

        if name not in component_cls._state_properties_:
            raise AttributeError

        state_property = getattr(component_cls, name)
        return PleaseTurnThisIntoAnAttributeBinding(state_property)


class PleaseTurnThisIntoAnAttributeBinding:
    def __init__(self, state_property: StateProperty):
        self.state_property = state_property



import types
from dataclasses import dataclass
from typing import *  # type: ignore

import introspection.typing

__all__ = ["ScopedAnnotation"]


@dataclass
class ScopedAnnotation:
    annotation: introspection.types.TypeAnnotation
    module: types.ModuleType

    def evaluate(self) -> introspection.types.TypeAnnotation:
        return introspection.typing.resolve_forward_refs(
            self.annotation,
            self.module,
            mode="eval",
            strict=False,
            treat_name_errors_as_imports=True,
        )



from __future__ import annotations

import collections
import functools
import sys
from collections.abc import Collection, Iterator, Mapping

import introspection.typing

import rio

__all__ = [
    "get_local_annotations",
    "get_resolved_type_annotations",
    "get_child_component_containing_attribute_names",
    "get_child_component_containing_attribute_names_for_builtin_components",
]


# Note: This function purposely isn't cached because calling it at different
# times can give different outputs. For example, if called immediately after the
# input class has been created, some forward references may not be evaluatable
# yet.
class get_local_annotations(Mapping[str, introspection.types.TypeAnnotation]):
    def __init__(self, cls: type, *, strict: bool = False):
        # Note: Don't use `typing.get_type_hints` because it has a stupid bug in
        # python 3.10 where it dies if something is annotated as
        # `dataclasses.KW_ONLY`.
        self._annotations: dict = vars(cls).get("__annotations__", {})
        self._module = sys.modules[cls.__module__]
        self._strict = strict

    def __getitem__(self, name: str) -> introspection.types.TypeAnnotation:
        return introspection.typing.resolve_forward_refs(
            self._annotations[name],
            self._module,
            mode="ast",
            strict=self._strict,
            treat_name_errors_as_imports=True,
        )

    def __iter__(self) -> Iterator[str]:
        return iter(self._annotations)

    def __len__(self) -> int:
        return len(self._annotations)


def get_resolved_type_annotations(
    cls: type,
) -> Mapping[str, type]:
    maps = [get_local_annotations(c, strict=True) for c in cls.__mro__]
    return collections.ChainMap(*maps)  # type: ignore


@functools.lru_cache(maxsize=None)
def get_child_component_containing_attribute_names(
    cls: type[rio.Component],
) -> Collection[str]:
    from . import serialization

    attr_names: list[str] = []

    for attr_name, serializer in serialization.get_attribute_serializers(
        cls
    ).items():
        # : Component
        if serializer is serialization._serialize_child_component:
            attr_names.append(attr_name)
        elif isinstance(serializer, functools.partial):
            # : Component | None
            if (
                serializer.func is serialization._serialize_optional
                and serializer.keywords["serializer"]
                is serialization._serialize_child_component
            ):
                attr_names.append(attr_name)
            # : list[Component]
            elif (
                serializer.func is serialization._serialize_list
                and serializer.keywords["item_serializer"]
                is serialization._serialize_child_component
            ):
                attr_names.append(attr_name)

    return tuple(attr_names)


@functools.lru_cache(maxsize=None)
def get_child_component_containing_attribute_names_for_builtin_components() -> (
    Mapping[str, Collection[str]]
):
    from .components.fundamental_component import FundamentalComponent

    result = {
        cls._unique_id: get_child_component_containing_attribute_names(cls)
        for cls in introspection.iter_subclasses(FundamentalComponent)
        if cls._unique_id.endswith("-builtin")
    }

    result.update(
        {
            "Placeholder": ["_child_"],
            "Align-builtin": ["child"],
            "Margin-builtin": ["child"],
        }
    )
    return result
